/*
 *  ORFA -- Open Robotics Firmware Architecture
 *
 *  Copyright (c) 2009-2010 Vladimir Ermakov, Andrey Demenev, Anton Botov
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the "Software"), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 *****************************************************************************/

#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>

#include <ctype.h>
#include "parser.h"
#include "lib/hex.h"

#ifndef PROTOCOL_VERSION_STRING
# ifdef SG_PROTOCOL_V1_0
#  define PROTOCOL_VERSION_STRING  "1.0"
# else
#  define PROTOCOL_VERSION_STRING  "1.1"
# endif
#endif

//! command buffer
static cbf_t tx_buf;

static CMD_PARSER* cmdparsers;

void register_parser(char cmd,LTR_PARSER* parser){
	CMD_PARSER x;
	x.parser=parser;
	x.next=cmdparsers;
	x.cmd=cmd;
	cmdparsers=&x;
}

void print_error(error_code_t error_code)
{
	putchar('E');
	putchar('R');
	putchar('R');
	putchar('O');
	putchar('R');
	putchar(' ');
	putchar(itox(error_code >> 4));
	putchar(itox(error_code & 0x0f));
	putchar('\n');
}

/*
 * Execute a cmd_buf sequence.
 *
 * \param[in]  cmd_buf buffer conraining the output command string.
 * \param[out] tx_buf buffer contining the messages sent to the user.
 * \return true on success, false on error.
 */
bool exec_cmd(cbf_t *cmd_buf, cbf_t *tx_buf, error_code_t *error_code)
{
	uint8_t c;
	uint8_t nbytes;
	uint16_t freq;
	uint8_t addr=0xaa;

	if (cbf_isempty(cmd_buf)) {
		return true;
	}
	c = cbf_get(cmd_buf);

	switch (c) {
		case 'V':
			cbf_put(tx_buf, c);
			for (uint8_t i = 0; i < sizeof(PROTOCOL_VERSION_STRING) - 1; i++) {
				cbf_put(tx_buf, PROTOCOL_VERSION_STRING[i]);
			}
			break;

		case 'C':			
			if (!cbf_isempty(cmd_buf)) {
				freq = (cbf_get(cmd_buf) << 8) | cbf_get(cmd_buf);
				if ((freq > 0) && (freq <= 800)) {
					i2c_config(freq);
				}
			}
			freq = i2c_get_freq();

			cbf_put(tx_buf, 'C');
			cbf_put(tx_buf, itox((freq >> 12) & 0x0f));
			cbf_put(tx_buf, itox((freq >>  8) & 0x0f));
			cbf_put(tx_buf, itox((freq >>  4) & 0x0f));
			cbf_put(tx_buf, itox((freq >>  0) & 0x0f));
			break;

		case 'L':
			if (!cbf_isempty(cmd_buf)) {
				addr = cbf_get(cmd_buf) & (~0x01);
				i2c_set_localhost(addr);
			}
			addr = i2c_get_localhost();

			cbf_put(tx_buf, 'L');
			cbf_put(tx_buf, itox((addr >>  4) & 0x0f));
			cbf_put(tx_buf, itox((addr >>  0) & 0x0f));
			break;

		case 'S':
			// I2C Read/Write command
			do {
				cbf_put(tx_buf, 'S');
				c = cbf_get(cmd_buf);
				if (c & 0x01) {
					// Read
					if (!i2c_master_start(c, i2c_rd)) {
						i2c_master_stop();
						debug("# ERROR: Nack on rd address\n");
						*error_code = NACK_ADDRESS;
						return false;
					}
					cbf_put(tx_buf, 'R');

					nbytes = cbf_get(cmd_buf);
					while (nbytes > 0) {
						if (!i2c_master_rxc(&c, nbytes > 1)) {
							i2c_master_stop();
							debug("# ERROR: Nack on rd byte\n");
							*error_code = NACK_BYTE;
							return false;
						}

						cbf_put(tx_buf, itox(c >> 4));
						cbf_put(tx_buf, itox(c & 0x0f));
						nbytes--;
					}
					c = cbf_get(cmd_buf);

				} else {
					// Write
					if (!i2c_master_start(c, i2c_wr)) {
						i2c_master_stop();
						debug("# ERROR: Nack on wr address\n");
						*error_code = NACK_ADDRESS;
						return false;
					}
					cbf_put(tx_buf, 'W');

					c = cbf_get(cmd_buf);
					while (c != 'S' && c != 'P') {
						nbytes = c;
						while (nbytes > 0) {
							c = cbf_get(cmd_buf);
							if (i2c_master_txc(c)) {
								cbf_put(tx_buf, 'A');
							} else {
								i2c_master_stop();
								debug("# ERROR: Nack on wr byte\n");
								*error_code = NACK_BYTE;
								return false;
							}
							nbytes--;
						}
						c = cbf_get(cmd_buf);
					}
				}
			} while (c == 'S');

			i2c_master_stop();

			if (c != 'P') {
				debug("# ERROR: P expected\n");
				*error_code = P_EXPECTED;
				return false;
			}

			cbf_put(tx_buf, 'P');
			break;

		case 'X':
			i2c_bus_clear();
			cbf_put(tx_buf, 'X');
			break;

		default:
			debug("# exec_cmd() internal error\n");
			return false;
			break;
	}

	// not needed: serial_fdev translate '\n' to '\r\n'
	// cbf_put(tx_buf, '\r');

	cbf_put(tx_buf, '\n');

	return true;
}

/**
 * Get one byte state machine
 * @param[in] c char
 * @param[out] *data received byte
 * @param[out] *error_code error reporting
 * @return true when receive complete
 */
static inline bool get_cmd_byte(uint8_t c, uint8_t *data, error_code_t *error_code)
{
	static uint8_t count = 1;

	int8_t tmp = xtoi(c);
	if (tmp >= 0 && count == 1) {
		// get first nibble
		*data = tmp << 4;
		count = 2;
	} else if (tmp >= 0 && count == 2) {
		// get second nibble
		*data |= tmp;
		count = 1;
		return true;
	} else if (c == ' ' && count == 1) {
		// skip spaces
	} else {
		count = 1;
		*error_code = INVALID_XDIGIT;
	}

	return false;
}

bool v_parser(uint8_t ch, cbf_t *cmd_buf, cbf_t *tx_buf, error_code_t *error_code, bool init, bool exec)
{
	cbf_put(*tx_buf, c);
	for (uint8_t i = 0; i < sizeof(PROTOCOL_VERSION_STRING) - 1; i++) {
		cbf_put(*tx_buf, PROTOCOL_VERSION_STRING[i]);
	};
	return true; //finished!
};

bool x_parser(uint8_t ch, cbf_t *cmd_buf, cbf_t *tx_buf, error_code_t *error_code, bool init, bool exec)
{
	i2c_bus_clear();
	cbf_put(*tx_buf, 'X');
	return true; //finished!
};

bool l_parser(uint8_t ch, cbf_t *cmd_buf, cbf_t *tx_buf, error_code_t *error_code, bool init, bool exec)
{
	if(exec){
		if (!cbf_isempty(cmd_buf)) {
			addr = cbf_get(cmd_buf) & (~0x01);
			i2c_set_localhost(addr);
		}
		addr = i2c_get_localhost();
		cbf_put(tx_buf, 'L');
		cbf_put(tx_buf, itox((addr >>  4) & 0x0f));
		cbf_put(tx_buf, itox((addr >>  0) & 0x0f));
		return true;
	};

	if(init){
		return false; //more bytes need
	}else{
		if (get_cmd_byte(c, &data, &error_c)) {
			cbf_put(cmd_buf, data);
			return true; //finished!
		};
		return false; //more bytes need
	};
};


bool c_parser(uint8_t ch, cbf_t *cmd_buf, cbf_t *tx_buf, error_code_t *error_code, bool init, bool exec)
{
	if(exec){
		if (!cbf_isempty(cmd_buf)) {
			addr = cbf_get(cmd_buf) & (~0x01);
			i2c_set_localhost(addr);
		}
		addr = i2c_get_localhost();
		cbf_put(tx_buf, 'L');
		cbf_put(tx_buf, itox((addr >>  4) & 0x0f));
		cbf_put(tx_buf, itox((addr >>  0) & 0x0f));
		return true;
	};

	if(init){
		return false; //more bytes need
	}else{
		if (get_cmd_byte(c, &data, &error_c)) {
			cbf_put(cmd_buf, data);
			return true; //finished!
		};
		return false; //more bytes need
	};
};

void parser_init(void)
{
	cbf_init(&tx_buf);
	cmdparsers=NULL;
	register_parser('V',&v_parser);
	register_parser('X',&x_parser);
	register_parser('L',&l_parser);
	register_parser('C',&c_parser);
}

bool parse_cmd(uint8_t co, cbf_t *cmd_buf)
{
	//! error state
	error_code_t error_code=NO_ERROR;

	// temp
	int8_t i;
	uint8_t c=toupper(co);
	c = (c == '\r')? '\n' : c;

	// static per iteration data
	static uint8_t count=1;
	static uint8_t data=0;
	static uint8_t addr=0;
	static uint8_t reg_len=1;
	static error_code_t error_c=NO_ERROR;

	// machine states
	static state_cmd_t state_cmd=CMD_INIT;
	static state_i2creg_t state_i2creg=GET_ADDRESS;

	if (state_cmd == CMD_INIT) {
		// Clear machine
		addr = 0;
		count = 1;
		data = 0;
		reg_len = 1;
		cbf_init(cmd_buf);
		state_cmd = GET_COMMAND;
		state_i2creg = GET_ADDRESS;
		error_c = error_code;
	}

	switch(state_cmd) {
		case GET_COMMAND:
			switch(c) {
				case 'S':
					// is i2c?
					state_cmd = PARSE_I2C;
					cbf_put(cmd_buf, 'S');
					break;

				case 'C':
					cbf_put(cmd_buf, c);
					state_cmd = PARSE_CLOCK;
					break;

				case 'L':
					cbf_put(cmd_buf, c);
					state_cmd = PARSE_LOCAL;
					break;

				case 'V':
				case 'X':
					cbf_put(cmd_buf, c);
					state_cmd = WAIT_EOL;
					break;

				case ' ':
				case '\n':
					break;

				case '%':
					// is comment?
					state_cmd = COMMENT_OR_ERROR_EOL;
					break;

				default:
					error_c = INVALID_COMMAND;
					break;
			}
			break;

		case COMMENT_OR_ERROR_EOL:
			// skip line
			if (c == '\n') {
				error_code = error_c;
				state_cmd = CMD_INIT;
			}
			break;

		case PARSE_CLOCK:
			if (get_cmd_byte(c, &data, &error_c)) {
				cbf_put(cmd_buf, data);
				if (count++ > 1) {
					state_cmd = WAIT_EOL;
				}
			}
			break;

		case PARSE_LOCAL:
			if (get_cmd_byte(c, &data, &error_c)) {
				cbf_put(cmd_buf, data);
				state_cmd = WAIT_EOL;
			}
			break;

		case PARSE_I2C:
			switch(state_i2creg) {
				case GET_ADDRESS:
					if (get_cmd_byte(c, &data, &error_c)) {
						cbf_put(cmd_buf, data);
						state_i2creg = (data & 0x01) ? PARSE_I2C_READ : PARSE_I2C_WRITE;
					}
					break;

				case PARSE_I2C_READ:
					if (get_cmd_byte(c, &data, &error_c)) {
						cbf_put(cmd_buf, data);
						state_i2creg = WAIT_RESTART_OR_STOP;
					}
					break;

				case WAIT_RESTART_OR_STOP:
					switch(c) {
						case 'S':
							// reSTART?
							cbf_put(cmd_buf, c);
							state_i2creg = GET_ADDRESS;
							break;

						case 'P':
							// STOP?
							cbf_put(cmd_buf, c);
							state_cmd = WAIT_EOL;
							break;

						default:
							break;
					}
					break;

				case PARSE_I2C_WRITE:
					i = xtoi(c);
					if (i >= 0 && count == 1) {
						// get first nibble
						data = i << 4;
						count = 2;
					} else if (i >= 0 && count == 2) {
						// get second nibble
						data |= i;
						cbf_put(cmd_buf, 1);
						cbf_put(cmd_buf, data);
						count = 1;
					} else if (c == 'S' && count == 1) {
						// reSTART?
						cbf_put(cmd_buf, c);
						state_i2creg = GET_ADDRESS;
					} else if (c == 'P' && count == 1) {
						// STOP?
						cbf_put(cmd_buf, c);
						state_cmd = WAIT_EOL;
					} else if (c == ' ' && count == 1) {
						// skip spaces
					} else {
						error_c = INVALID_DATA;
					}
					break;

				default:
					break;
			} // parse_i2c state machine
			break;


		case WAIT_EOL:
			if (c == '\n') {
				// EOL? (exec)
				cbf_put(cmd_buf, '\n');
				state_cmd = EXEC_COMMAND;
			}
			break;

		default:
			error_code = INTERNAL_ERROR;
			break;
	}

	// ---------------------------------------------------------------

	if (error_c != NO_ERROR) {
		state_cmd = COMMENT_OR_ERROR_EOL;
	}

	if (error_code != NO_ERROR) {
		state_cmd = CMD_INIT;
	}

	if (state_cmd == EXEC_COMMAND) {
		state_cmd = CMD_INIT;
		if(exec_cmd(cmd_buf, &tx_buf, &error_code))
		{
			while(!cbf_isempty(&tx_buf))
			{
				putchar(cbf_get(&tx_buf));
			}
		}
	}

	if(error_code != NO_ERROR)
	{
		print_error(error_code);
		cbf_init(&tx_buf);
		error_code = NO_ERROR;
	}

	return false;
}

